/***************************************************************************/
/* JDE driver example.                                                     */
/* It is a driver that controls a 'null actuator' wich only has one input  */
/* variable. It just prints on the screen the value of this input variable */
/* ----------------------------------------------------------------------  */
/* NULL DRIVER API:                                                        */
/*    -value: The actuator input variable                                  */
/***************************************************************************/
#include <stdio.h>
#include <pthread.h>
#include <math.h>
#include "jde.h"
#include "serial.h"

/*-- HEADERS -- */
#define TRAMA_ENA_CAB 0x45
#define TRAMA_POS_CAB 0x57


/* ----------------------------- DRIVER INTERFACE ------------------------- */
#define NAME "servos8"

typedef struct  {
  char pos[8];     /* Servo angular position (in degrees) [-90,90] */
  char enable[8];  /* Servo enabled/disabled                       */
} drv_interface;

drv_interface interface;
/*--------------------------------------------------------------------------*/



/* servos8 driver main pthread.*/
pthread_t servos8_th;

/* servos8 driver variable to detect if pthreads must end its execution.*/
int servos8_close_command=0;

/* servos8 driver state variable for the actuator pthread.*/
int state_actuator;

/* servos8 driver mutex for the actuator pthread.*/
pthread_mutex_t mymutex_actuator;

/* servos8 driver condition flag for the actuator pthread.*/
pthread_cond_t condition_actuator;

/* 'servos8' schema cycle control variable*/
int servos8_cycle=30;

/* id for servos8 actuator schema.*/
int servos8_schema_id;

/* servos8 ref counter*/
int servos8_refs=0;

/* mutex for ref counters*/
pthread_mutex_t refmutex;

/* Serial port descriptor */
int serial_fd;           

/************************************************/
/* SERIAL COMUNICATIONS                         */
/************************************************/
void sg_servos8_pos1(int servo, int pos)
{
	char trama[4];
  
  pos = (int)round(69 + (23.0/30.0)*(float)pos);

  /*pos =(-1.28*pos + 115);*/
  
  /* Construir la trama POS */
  trama[0]=TRAMA_POS_CAB;
	trama[1]=(char)servo;
	trama[2]=(char)pos;

  /* -- Enviarla al StarGate */
  serial_send(serial_fd, trama,3);
  
}

void sg_servos8_enable(int mask)
{
	char trama[3];

  /* Construir la trama ENA */
  trama[0]=TRAMA_ENA_CAB;
	trama[1]=(char)mask;

  /*-- Enviarla al StarGate */
  serial_send(serial_fd, trama,2);
}

/* Entry point: servos8 driver function to stop and clean servos8 devices */
void servos8_close()
{
  servos8_close_command=1;
  serial_close(serial_fd);
  printf("driver servos8 off\n");
}

/* Entry point: suspend the servos8 device */
int servos8_suspend()
{
   pthread_mutex_lock(&refmutex);
   if (servos8_refs>1){
      servos8_refs--;
      pthread_mutex_unlock(&refmutex);
   }
   else{
      servos8_refs=0;
      pthread_mutex_unlock(&refmutex);
      
         pthread_mutex_lock(&mymutex_actuator);
         state_actuator=slept;
         put_state(servos8_schema_id,slept);
         printf("servos8 schema suspend\n");
         sg_servos8_enable(0x00);
         pthread_mutex_unlock(&mymutex_actuator);
   }
   return 0;
}

/* Entry point: resume the servos8 device */
int servos8_resume(int father, int *brothers, arbitration fn)
{
   sg_servos8_enable(0x01);
  
   pthread_mutex_lock(&refmutex);
   if (servos8_refs>0){
      servos8_refs++;
      pthread_mutex_unlock(&refmutex);
   }
   else{
      servos8_refs=1;
      pthread_mutex_unlock(&refmutex);

      pthread_mutex_lock(&mymutex_actuator);
      state_actuator=winner;
      all[servos8_schema_id].father = father;
      all[servos8_schema_id].fps = 0.;
      all[servos8_schema_id].k =0;
      put_state(servos8_schema_id,winner);
      printf("servos8 schema resume\n");
      pthread_cond_signal(&condition_actuator);
      pthread_mutex_unlock(&mymutex_actuator);
   }
   return 0;
}

/* Perform an interation on the servos8 device */
void servos8_iteration()
{  
  static char old_pos[8]={0,0,0,0,0,0,0,0};
  static int old_enable[8]={0,0,0,0,0,0,0,0};
  char msg[2];
  int i;
  
  speedcounter(servos8_schema_id);
  
  for (i=0; i<8; i++) {
    if (interface.enable[i]!=old_enable[i]) {
      printf ("ENABLE!\n");
      if (interface.enable[i]) 
        sg_servos8_enable(0xFF);
      else 
        sg_servos8_enable(0x00);
      
    }
   
    /* Print the actuador input variable */
    /* Only if it has changed!           */
    if (interface.pos[i]!=old_pos[i]) {
      msg[0]=interface.pos[i];
      printf("Servo %i pos: %d\n",i+1,interface.pos[i]);
      /*-- Send the command to the serial port */
      sg_servos8_pos1(i+1, interface.pos[i]);
    }  
  
    old_pos[i]=interface.pos[i];
    old_enable[i]=interface.enable[i];
  }  
}

/* Nullactuator main thread */
void *servos8_thread(void *not_used) 
{
  struct timeval a,b;
  long diff, next;

  do{
      
    if (state_actuator==slept) {
	    pthread_mutex_lock(&mymutex_actuator);
	    pthread_cond_wait(&condition_actuator,&mymutex_actuator);
	    pthread_mutex_unlock(&mymutex_actuator);
	  }
	  else  {
	    pthread_mutex_unlock(&mymutex_actuator);
	    gettimeofday(&a,NULL);
	    servos8_iteration();
	    gettimeofday(&b,NULL);  
	    diff = (b.tv_sec-a.tv_sec)*1000000+b.tv_usec-a.tv_usec;
	    
	    next = servos8_cycle*1000-diff-10000; 
	    /* discounts 10ms taken by calling usleep itself */
	    if (next>0) usleep(servos8_cycle*1000-diff);
	    else {
	      printf("time interval violated: servos8\n");
	      usleep(servos8_cycle*1000);
	    }
	  }
      
  }while(servos8_close_command==0);

  pthread_exit(0);
}

void servos8_init()
{
  /*-- Open the serial port                */
  /*-- The speed is configure at 9600 baud */
  serial_fd=serial_open("/dev/ttyUSB0",B9600);
  
}

/* Entry point: Start the driver and the servos8 actuator device */
void servos8_startup(char *configfile)
{
  servos8_init();
  interface.pos[0]=0;
  interface.enable[0]=0;
  
  
  pthread_mutex_lock(&mymutex_actuator);
  state_actuator=slept;
  pthread_create(&servos8_th,NULL,servos8_thread,NULL);
  pthread_mutex_unlock(&mymutex_actuator);    

  all[num_schemas].id = (int *) &servos8_schema_id;
  strcpy(all[num_schemas].name,NAME);
  all[num_schemas].resume = (resumeFn) servos8_resume;
  all[num_schemas].suspend = (suspendFn) servos8_suspend;
  printf("%s schema loaded (id %d)\n",all[num_schemas].name,num_schemas);
  (*(all[num_schemas].id)) = num_schemas;
  all[num_schemas].fps = 0.;
  all[num_schemas].k =0;
  all[num_schemas].state=slept;
  all[num_schemas].close = NULL;
  all[num_schemas].handle = NULL;
  num_schemas++;
  
  myexport(NAME, "interface", &interface);
  myexport(NAME,"id",&servos8_schema_id);
  myexport(NAME,"cycle", &servos8_cycle);
  myexport(NAME,"resume",(void *)&servos8_resume);
  myexport(NAME,"suspend",(void *)&servos8_suspend);

  printf("servos8 driver started up\n");
}


